#include <actions.h>
#include <gears.h>
#include <debug.h>
#include <string.h>
#include <browser.h>
#include <packagekit/pk-enum.h>

extern Ecore_Evas * ee;
struct Installer_Progressbar *_AP_;

static char * progress_bar_type_text (enum ActionType type) {
	if (type == INSTALL ||
	        type == UPDATE) {
		return "Downloading ";
	} else if (type == UNINSTALL) {
		return "Uninstalling ";

	} else {
		return "";
	}
}

static void pk_update_progress_cb (void *data, DBusMessage *message);
static void pk_change_status_cb (void *data, DBusMessage *message);
static void pk_error_cb (void *data, DBusMessage *message);
static void pk_message_cb (void *data, DBusMessage *message);
static void __delayed_fade_out(void *data, DBusMessage *message);

void action(struct Target *target) {
	Pk_Operation * pko;
	Evas *evas = ecore_evas_get(ee);

	if (!target) {
		error("What is your target!!\n");
		return;
	}
	pko = (Pk_Operation *)malloc(sizeof(Pk_Operation));
	memset(pko, 0, sizeof(Pk_Operation));

	pko->payload = target;

	// Adding general call back here
	pko->Operations[ProgressChanged_sig] = 	&pk_update_progress_cb;
	pko->Operations[StatusChanged_sig] = 	&pk_change_status_cb;
	pko->Operations[ErrorCode_sig] = 	&pk_error_cb;
	pko->Operations[Message_sig] = 		&pk_message_cb;
	pko->Operations[Finished_sig] = 	&__delayed_fade_out;



	_AP_ = new_Progressbar(evas, progress_bar_type_text(target->type));

	if (target->type == INSTALL) {
		debug("Do install!\n");
		pk_install_package(pko, target->package);
	} else if (target->type == UPDATE) {
		debug("Do update!\n");
		pk_update_package(pko, target->package);
	} else if (target->type == UNINSTALL) {
		debug("Do uninstall!\n");
		pk_uninstall_package(pko, target->package);
	} else {
		error("Hey!! Bad type %d!!\n", target->type);
		free_Progressbar(_AP_);
		free(pko);
		_AP_ = NULL;
		return;
	}
}


static void pk_update_progress_cb (void *data, DBusMessage *message) {
	Pk_Operation * pko = (Pk_Operation *)data;
	struct Target * target = (struct Target *)pko->payload;
	DBusError _err;
	char *tid;
	unsigned int percentage, subpercentage, elapsed, remaining;
	debug("get progress cb!!\n");
	dbus_error_init(&_err);
	if (!dbus_message_get_args(message, &_err,
	                           DBUS_TYPE_STRING, &tid,
	                           DBUS_TYPE_UINT32, &percentage,
	                           DBUS_TYPE_UINT32, &subpercentage,
	                           DBUS_TYPE_UINT32, &elapsed,
	                           DBUS_TYPE_UINT32, &remaining,
	                           DBUS_TYPE_INVALID)) {
		error("Error Message: %s\n", _err.message);
		dbus_error_free(&_err);
		return;
	}
	dbus_error_free(&_err);
	if (!pk_should_proxy(pko, tid)) {
		return;
	}
	target->progress = percentage;
	debug("Progress: %d %d %d %d\n", percentage, subpercentage, elapsed, remaining);
	if (_AP_) {
		if ( target->status == DOWNLOADING) {
			progressbar_update(_AP_, target->progress);
		} else if ( target->status == INITIAL ) {
			// Do nothing
		} else {
		}
	}
}

static int progressbar_delay_die(void *data) {
	if (_AP_) {
		free_Progressbar(_AP_);
		_AP_ = NULL;
	}
	scroll_page_back_to_root();
	return 0;
}

//pk_marshal_VOID__STRING_STRING change_status
static void pk_change_status_cb (void *data, DBusMessage *message) {
	char *tid;
	char *status_text;
	Pk_Operation * pko = (Pk_Operation *)data;
	struct Target * target = (struct Target *)pko->payload;
	DBusError _err;
	dbus_error_init(&_err);
	if (!dbus_message_get_args(message, &_err,
	                           DBUS_TYPE_STRING, &tid,
	                           DBUS_TYPE_STRING, &status_text,
	                           DBUS_TYPE_INVALID )) {
		error("Error Message: '%s'\n", _err.message);
		dbus_error_free(&_err);
		return;
	}
	dbus_error_free(&_err);
	if (!pk_should_proxy(pko, tid)) {
		return;
	}
	debug("Status Changed: %s\n", status_text);
	if (!_AP_) {
		error("_AP_ is NULL!!\n");
		return;
	}
	if (pk_status_enum_from_text(status_text) == PK_STATUS_ENUM_FINISHED) {
		if (target->status != FAILED) {
			target->status = SUCCESS;
		}
		progressbar_update(_AP_, 100);
		if (target->status == SUCCESS) {
			progressbar_text_set(_AP_, "Success...");
		} else {
			progressbar_text_set(_AP_, "Failed...");
		}
	} else if (target->type == INSTALL) {
		if (pk_status_enum_from_text(status_text) == PK_STATUS_ENUM_DOWNLOAD) {
			target->status = DOWNLOADING;
			progressbar_text_set(_AP_, "Downloading ");
		} else if (pk_status_enum_from_text(status_text) == PK_STATUS_ENUM_INSTALL) {
			target->status = INSTALLING;
			progressbar_update(_AP_, 100);
			progressbar_text_set(_AP_, "Processing...");
		} else if (pk_status_enum_from_text(status_text) == PK_STATUS_ENUM_SETUP) {
			target->status = INITIAL;
			progressbar_update(_AP_, 0);
			progressbar_text_set(_AP_, "Preparing...");
		} else {
			error("Hun? %s!!\n", status_text);
		}
	} else if (target->type == UPDATE) {
		if (pk_status_enum_from_text(status_text) == PK_STATUS_ENUM_DOWNLOAD) {
			target->status = DOWNLOADING;
			progressbar_text_set(_AP_, "Downloading ");
		} else if (pk_status_enum_from_text(status_text) == PK_STATUS_ENUM_UPDATE) {
			target->status = UPDATING;
			progressbar_update(_AP_, 0);
			progressbar_text_set(_AP_, "Updating...");
		} else if (pk_status_enum_from_text(status_text) == PK_STATUS_ENUM_SETUP) {
			target->status = INITIAL;
			progressbar_update(_AP_, 0);
			progressbar_text_set(_AP_, "Preparing...");
		} else {
			error("Hun? %s!!\n", status_text);
		}
	} else if (target->type == UNINSTALL) {
		progressbar_update(_AP_, 100);
	}
}


static void pk_error_cb (void *data, DBusMessage *message) {
	char *tid;
	unsigned int error_code;
	char *detail;
	Pk_Operation * pko = (Pk_Operation *)data;
	DBusError _err;
	dbus_error_init(&_err);
	if (!dbus_message_get_args(message, &_err,
	                           DBUS_TYPE_STRING, &tid,
	                           DBUS_TYPE_UINT32, &error_code,
	                           DBUS_TYPE_STRING, &detail,
	                           DBUS_TYPE_INVALID
	                          )) {
		error("Error Message: '%s'\n", _err.message);
		dbus_error_free(&_err);
		return;
	}
	dbus_error_free(&_err);
	if (!pk_should_proxy(pko, tid)) {
		return;
	}
	debug("error_code is %d '%s'\n", error_code, detail);
}

static void pk_message_cb (void *data, DBusMessage *message) {
	char *tid;
	unsigned int message_code;
	char *detail;
	Pk_Operation * pko = (Pk_Operation *)data;
	DBusError _err;
	dbus_error_init(&_err);
	if (!dbus_message_get_args(message, &_err,
	                           DBUS_TYPE_STRING, &tid,
	                           DBUS_TYPE_UINT32, &message_code,
	                           DBUS_TYPE_STRING, &detail,
	                           DBUS_TYPE_INVALID
	                          )) {
		error("Error Message: '%s'\n", _err.message);
		dbus_error_free(&_err);
		return;
	}
	dbus_error_free(&_err);
	if (!pk_should_proxy(pko, tid)) {
		return;
	}
	debug("message_code is %d '%s'\n", message_code, detail);
}

static void __delayed_fade_out(void *data, DBusMessage *message) {
	char *tid;
	char *exit_text;
	unsigned int runtime;
	Pk_Operation * pko = (Pk_Operation *)data;
	DBusError _err;
	dbus_error_init(&_err);
	if (!dbus_message_get_args(message, &_err,
	                           DBUS_TYPE_STRING, &tid,
	                           DBUS_TYPE_STRING, &exit_text,
	                           DBUS_TYPE_UINT32, &runtime,
	                           DBUS_TYPE_INVALID )) {
		error("Error Message: '%s'\n", _err.message);
		dbus_error_free(&_err);
		return;
	}
	dbus_error_free(&_err);
	if (!pk_should_proxy(pko, tid)) {
		return;
	}
	debug("Transaction '%s' is finished!\n", pko->tid);
	debug("exit_text: %s\n", exit_text);
	debug("runtime %d\n", runtime);

	ecore_timer_add(3, &progressbar_delay_die, NULL);

	pk_default_get_finished_cb(pko, message);

	refresh_categories_lists();
	return;
}
