#include <pkutil.h>
#include <debug.h>
#include <string.h>
#include <stdlib.h>
#include <pk-package-id.h>
static E_DBus_Connection* e_conn = NULL;
static E_DBus_Proxy *__e_proxy = NULL;

Pk_Operation * starts_a_new_translation(Pk_Operation * pko, const char *tid);


const char * find_keyword(Operations_CB num) {
	static const char * __POM [] = {
		"ProgressChanged",
		"StatusChanged",
		"Finished",
		"ErrorCode",
		"Message",
		"Description",
		"Files",
		"Package",
		"RepoDetail",
		"UpdateDetail",
		NULL,
	};
	if (num < 0 || num >= END_OF_CB_SIG) {
		return NULL;
	}
	return __POM[num];
}

E_DBus_Connection * pk_e_dbus_init(void) {
	if (e_conn) return e_conn;

	debug("E_DBUS initialing. \n");
	if (!e_dbus_init()) {
		return NULL;
	}

	e_conn = e_dbus_bus_get(PK_DBUS_BUS);

	if (!e_conn) {
		error("Get %d bus failed!!\n", PK_DBUS_BUS);
		pk_e_dbus_shutdown();
		return NULL;
	}
	return e_conn;
}

void pk_e_dbus_shutdown(void) {
	if (e_conn) {
		if (pk_proxy()) {
			e_dbus_proxy_destroy(pk_proxy());
			__e_proxy = NULL;
		}
		e_dbus_connection_close(e_conn);
		e_conn = NULL;
		e_dbus_shutdown();
	}
}

E_DBus_Proxy * pk_proxy(void) {
	if (__e_proxy) return __e_proxy;
	if (!pk_e_dbus_init()) {
		return NULL;
	}
	__e_proxy = e_dbus_proxy_new_for_name (pk_e_dbus_init(), PK_DBUS_SERVICE, PK_DBUS_PATH, PK_DBUS_INTERFACE);
	return __e_proxy;
}

int pk_should_proxy(Pk_Operation * pko, char *tid) {
	if (pko == NULL || pko->tid == NULL || tid == NULL) {
		error("bad arguments!!\n");
		return 0;
	}
	if (strncmp(pko->tid, tid, strlen(pko->tid))) {
		//debug("data=%s, tid:%s is different, Ignore this event!\n",pko->tid,tid);
		return 0;
	}
	return 1;
}

// pk_marshal_VOID__STRING_STRING_STRING
static void pk_default_message_cb (void *data, DBusMessage *message) {
	char *tid;
	char *_message;
	char *detail;
	Pk_Operation * pko = (Pk_Operation *)data;
	DBusError _err;
	dbus_error_init(&_err);
	if (!dbus_message_get_args(message, &_err,
	                           DBUS_TYPE_STRING, &tid,
	                           DBUS_TYPE_STRING, &_message,
	                           DBUS_TYPE_STRING, &detail,
	                           DBUS_TYPE_INVALID )) {
		error("Error Message: '%s'\n", _err.message);
		dbus_error_free(&_err);
		return;
	}
	dbus_error_free(&_err);
	if (!pk_should_proxy(pko, tid)) {
		return;
	}
	debug("Transcation: %s Message '%s' Detail: '%s'\n", tid, _message, detail);
}

// pk_marshal_VOID__STRING_STRING
static void pk_default_status_change_cb(void *data, DBusMessage *message) {
	char *tid;
	char *status_text;
	Pk_Operation * pko = (Pk_Operation *)data;
	DBusError _err;
	dbus_error_init(&_err);
	if (!dbus_message_get_args(message, &_err,
	                           DBUS_TYPE_STRING, &tid,
	                           DBUS_TYPE_STRING, &status_text,
	                           DBUS_TYPE_INVALID )) {
		error("Error Message: '%s'\n", _err.message);
		dbus_error_free(&_err);
		return;
	}
	dbus_error_free(&_err);
	if (!pk_should_proxy(pko, tid)) {
		return;
	}
	debug("Transcation: %s Status chages to %s\n", tid, status_text);
}

// pk_marshal_VOID__STRING_STRING_UINT  Finished
void pk_default_get_finished_cb (void *data, DBusMessage *message) {
	char *tid;
	char *exit_text;
	unsigned int runtime;
	int i;
	Pk_Operation * pko = (Pk_Operation *)data;
	DBusError _err;
	dbus_error_init(&_err);
	if (!dbus_message_get_args(message, &_err,
	                           DBUS_TYPE_STRING, &tid,
	                           DBUS_TYPE_STRING, &exit_text,
	                           DBUS_TYPE_UINT32,   &runtime,
	                           DBUS_TYPE_INVALID )) {
		error("Error Message: '%s'\n", _err.message);
		dbus_error_free(&_err);
		return;
	}
	dbus_error_free(&_err);
	if (!pk_should_proxy(pko, tid)) {
		return;
	}
	debug("Transaction '%s' is finished!\n", pko->tid);
	debug("exit_text: %s\n", exit_text);
	debug("runtime %d\n", runtime);
	for (i = 0;i < END_OF_CB_SIG;i++) {
		if (! pko->Operations[i] ) {
			continue;
		}
		e_dbus_proxy_disconnect_signal(pk_proxy(), find_keyword(i), pko->Operations[i], data);
	}
	free(pko->tid);
	memset(data, 0, sizeof(Pk_Operation));
	free(data);
	return;
}

Pk_Operation * starts_a_new_translation(Pk_Operation * pko, const char *tid) {
	int i;
	if (pko == NULL) {
		pko = (Pk_Operation *)malloc(sizeof(Pk_Operation));
		memset(pko, 0, sizeof(Pk_Operation));
	}
	if (!tid) {
		error("tid is NULL.\n");
		return NULL;
	}
	// Use default Finished event
	if (!pko->Operations[Finished_sig]) {
		pko->Operations[Finished_sig] 		= &pk_default_get_finished_cb;
	}
	if (!pko->Operations[StatusChanged_sig]) {
		pko->Operations[StatusChanged_sig] 	= &pk_default_status_change_cb;
	}
	if (!pko->Operations[Message_sig]) {
		pko->Operations[Message_sig] 		= &pk_default_message_cb;
	}
	if (pko->tid) {
		error("pko->tid shall be NULL now. contents: '%s'\n", pko->tid);
	}
	pko->tid = strdup(tid);
	for (i = 0;i < END_OF_CB_SIG ;i++) {
		if (!pko->Operations) {
			continue;
		}
		e_dbus_proxy_connect_signal(pk_proxy(), find_keyword(i), pko->Operations[i], pko);
	}
	return pko;
}

static char * pk_get_tid(char **tid) {
	DBusError _err;
	if (tid == NULL) return NULL;
	dbus_error_init(&_err);
	if (!pk_proxy()) {
		error("Get Proxy Failed!!\n");
		return NULL;
	}
	if (!e_dbus_proxy_simple_call(pk_proxy(), "GetTid", &_err,
	                              DBUS_TYPE_INVALID,
	                              DBUS_TYPE_STRING, tid,
	                              DBUS_TYPE_INVALID)) {
		error("Error message: %s\n", _err.message);
#ifdef CONFIG_DEBUG
		return NULL;
#else
		exit(0);
#endif

	}
	dbus_error_free(&_err);
	debug("Get tid %s\n", *tid);
	return *tid;
}


static void __pk_search(Pk_Operation * pko, const char *name, const char *what) {
	DBusError _err;
	int ret;
	char *filter;
	char *tid;
	if (!name || !what) {
		error("Search for what?\n");
		return;
	}
	if (!pk_get_tid(&tid)) {
		error("Get tid failed!!\n");
		return;
	}
	debug("tid=='%s' Action: %s\n", tid, what);
	dbus_error_init(&_err);
	filter = strdup(OMFILTER);
	starts_a_new_translation(pko, tid);
	ret = e_dbus_proxy_simple_call(pk_proxy(), what , &_err,
	                               DBUS_TYPE_STRING, &tid,
	                               DBUS_TYPE_STRING, &filter,
	                               DBUS_TYPE_STRING, &name,
	                               DBUS_TYPE_INVALID,
	                               DBUS_TYPE_INVALID);
	debug("Ret==%d\n", ret);
	if (ret == 0) {
		error("Error message: %s!!\n", _err.message);
	}
	free(filter);
	dbus_error_free(&_err);
}


void pk_search_name(Pk_Operation * pko, const char *name) {
	__pk_search(pko, name, "SearchName");
}



void pk_search_group(Pk_Operation * pko, const char *group) {
	__pk_search(pko, group, "SearchGroup");
}

static void __pk_get_updates(Pk_Operation * pko, const char *what) {
	DBusError _err;
	int ret;
	char *filter;
	char *tid;
	if (!what) {
		error("Search for what?\n");
		return;
	}
	if (!pk_get_tid(&tid)) {
		error("Get tid failed!!\n");
		return;
	}
	debug("tid=='%s' Action: %s\n", tid, what);
	dbus_error_init(&_err);
	filter = strdup(OMFILTER);
	starts_a_new_translation(pko, tid);
	ret = e_dbus_proxy_simple_call(pk_proxy(), what , &_err,
	                               DBUS_TYPE_STRING, &tid,
	                               DBUS_TYPE_STRING, &filter,
	                               DBUS_TYPE_INVALID,
	                               DBUS_TYPE_INVALID);
	debug("Ret==%d\n", ret);
	if (ret == 0) {
		error("Error message: %s!!\n", _err.message);
	}
	free(filter);
	dbus_error_free(&_err);
}


void pk_get_updates(Pk_Operation * pko) {
	__pk_get_updates(pko, "GetUpdates");
}

static void __pk_action_to_package(Pk_Operation * pko, const char *packageid, const char *action) {
	DBusError _err;
	int ret;
	char *tid;
	if (!packageid || !action) {
		error("Do what?\n");
		return;
	}
	if (!pk_get_tid(&tid)) {
		error("Get tid failed!!\n");
		return;
	}
	debug("tid=='%s'\n", tid);
	dbus_error_init(&_err);
	starts_a_new_translation(pko, tid);
	ret = e_dbus_proxy_simple_call(pk_proxy(), action , &_err,
	                               DBUS_TYPE_STRING, &tid,
	                               DBUS_TYPE_STRING, &packageid,
	                               DBUS_TYPE_INVALID,
	                               DBUS_TYPE_INVALID);
	debug("%s Ret==%d\n", action, ret);
	if (ret == 0) {
		error("Error message: %s!!\n", _err.message);
	}
	dbus_error_free(&_err);
}

void pk_get_description(Pk_Operation * pko, const char * packageid) {
	__pk_action_to_package(pko, packageid, "GetDescription");
}

void pk_install_package(Pk_Operation * pko, const char * packageid) {
	__pk_action_to_package(pko, packageid, "InstallPackage");
}

void pk_update_package(Pk_Operation * pko, const char * packageid) {
	__pk_action_to_package(pko, packageid, "UpdatePackage");
}


void _pk_uninstall_package(Pk_Operation * pko, const char * packageid, int allow_dep, int autoremove) {
	DBusError _err;
	int ret;
	char *tid;
	if (!packageid) {
		error("Remove what?\n");
		return;
	}
	if (!pk_get_tid(&tid)) {
		error("Get tid failed!!\n");
		return;
	}
	debug("tid=='%s'\n", tid);
	dbus_error_init(&_err);
	starts_a_new_translation(pko, tid);
	ret = e_dbus_proxy_simple_call(pk_proxy(), "RemovePackage" , &_err,
	                               DBUS_TYPE_STRING, &tid,
	                               DBUS_TYPE_STRING, &packageid,
	                               DBUS_TYPE_BOOLEAN, &allow_dep,
	                               DBUS_TYPE_BOOLEAN, &autoremove,
	                               DBUS_TYPE_INVALID,
	                               DBUS_TYPE_INVALID);
	debug("Ret==%d\n", ret);
	if (ret == 0) {
		error("Error message: %s!!\n", _err.message);
	}
	dbus_error_free(&_err);
}


void pk_uninstall_package(Pk_Operation * pko, const char * packageid) {
	// default turn on the auto install flag.
	_pk_uninstall_package(pko, packageid, 1, 1);
}

static void __pk_single_action(Pk_Operation * pko, const char *action) {
	DBusError _err;
	int ret;
	char *tid;
	if (!action) {
		error("Do what?\n");
		return;
	}
	if (!pk_get_tid(&tid)) {
		error("Get tid failed!!\n");
		return;
	}
	debug("tid=='%s'\n", tid);
	dbus_error_init(&_err);
	starts_a_new_translation(pko, tid);
	ret = e_dbus_proxy_simple_call(pk_proxy(), action , &_err,
	                               DBUS_TYPE_STRING, &tid,
	                               DBUS_TYPE_INVALID,
	                               DBUS_TYPE_INVALID);
	debug("Ret==%d\n", ret);
	if (ret == 0) {
		error("Error message: %s!!\n", _err.message);
	}
	dbus_error_free(&_err);
}

void pk_update_system(Pk_Operation * pko) {
	__pk_single_action(pko, "UpdateSystem");
}

static void __pk_refresh_action(Pk_Operation * pko, const char *action) {
	DBusError _err;
	int ret;
	int force=1;
	char *tid;
	if (!action) {
		error("Do what?\n");
		return;
	}
	if (!pk_get_tid(&tid)) {
		error("Get tid failed!!\n");
		return;
	}
	debug("tid=='%s'\n", tid);
	dbus_error_init(&_err);
	starts_a_new_translation(pko, tid);
	ret = e_dbus_proxy_simple_call(pk_proxy(), action , &_err,
	                               DBUS_TYPE_STRING, &tid,
				       DBUS_TYPE_BOOLEAN, &force,   // force
	                               DBUS_TYPE_INVALID,
	                               DBUS_TYPE_INVALID);
	debug("Ret==%d\n", ret);
	if (ret == 0) {
		error("Error message: %s!!\n", _err.message);
	}
	dbus_error_free(&_err);
}

// Do System database update such as opkg update
void pk_refresh_system(Pk_Operation * pko) {
	__pk_refresh_action(pko, "RefreshCache");
}


int get_pk_name_from_id (char *packageid, char *name, int bufsize) {
	PkPackageId * pid;
	if (!packageid || !name) {
		return 0;
	}
	if (!pk_package_id_check(packageid)) {
		if (packageid && strlen(packageid) > 0) {
			error("This packageid '%s' may be wrong!!\n", packageid);
			strncpy(name, packageid, bufsize);
			return 1;
		}
		error("Bad Packageid!!\n");
		name[0] = '\0';
		return 0;
	}
	pid	= pk_package_id_new_from_string(packageid);
	strncpy (name, pid->name, bufsize);
	pk_package_id_free(pid);
	return 1;
}

